from proto_formatter.comment import CommentParser
from proto_formatter.constant import Constant
from proto_formatter.proto_structures import Import
from proto_formatter.protobuf import Protobuf


class ImportParser(Constant):

    @classmethod
    def parse_and_add(cls, proto_obj: Protobuf, line, top_comment_list):
        new_import = cls.parse_import(line, top_comment_list)
        for existing_import in proto_obj.imports:
            if existing_import.value == new_import.value:
                raise Exception(f'multiple import detected: {new_import.value}!')

        proto_obj.imports.append(new_import)

    @classmethod
    def parse_import(cls, line, top_comment_list):
        value = cls._get_import_value(line)
        comments = CommentParser.create_comment(line, top_comment_list)
        n_import = Import(value, comments)
        return n_import

    @classmethod
    def _get_import_value(cls, line):
        line = line.strip()
        lindex = len('import ')
        rindex = line.index(cls.SEMICOLON)
        value = line[lindex:rindex].strip().replace("'", '').replace('"', '')
        return value



from proto_formatter.parser import ProtoParser


def format_file(fp, indents=2, top_comment=False, align_by_equal_sign=False, flatten=False, comment_max_length=None,
                new_fp=None):
    """
    Format protobuf file, override the original file if no new file path(new_fp) specified
    or write the formatted result to a new file.
    :param fp: unformatted protobuf file path.
    :param indents: indents number.
    :param top_comment: convert the single line comment at the left of code as a top comment or not. Default: not
    :param align_by_equal_sign: if align the code by equal sign or not.
    Example of align:
        ENUM_TYPE_UNSPECIFIED = 0;  // ENUM_TYPE_UNSPECIFIED
        ENUM_TYPE_CARRY_A     = 1;  // ENUM_TYPE_CARRY_A
        ENUM_TYPE_B           = 2;  // ENUM_TYPE_B
    :param flatten: if flatten nested object or not.
    Example of flatten:
      Original file:
         message Outer {
           int32 outerq = 1;
           message MiddleAA {
             message Inner {
               int64 ival = 1;
               bool  booly = 2;
             }
             bool  booly_aa = 1;
           }
         }

      Formatted file:
         message Inner {
             int64 ival = 1;
             bool booly = 2;
         }

         message MiddleAA {
             bool booly_aa = 1;
         }

         message Outer {
             int32 outerq = 1;
         }
    :param comment_max_length: the max length of comment line, defalut is no limitation.
    :param new_fp: the file path of new formatted protobuf file. Rewrite the original file if it is not specified.
    :return: file content size.
    """
    parser = ProtoParser()
    protobuf_obj = parser.load(fp=fp)
    content = protobuf_obj.to_string(
        indents=indents,
        top_comment=top_comment,
        align_by_equal_sign=align_by_equal_sign,
        flatten=flatten,
        comment_max_length=comment_max_length
    )

    if new_fp:
        fp = new_fp

    with open(fp, 'w') as f:
        return f.write(content)


def format_str(proto_str, indents=2, top_comment=False, align_by_equal_sign=False, flatten=False, comment_max_length=None):
    """
    Format a protobuf string, return the formatted string.
    :param proto_str: protobuf string need to be formatted.
    :param indents: indents number.
    :param top_comment: convert the single line comment at the left of code as a top comment or not. Default: not
    :param align_by_equal_sign: if align the code by equal sign or not.
    Example of align:
        ENUM_TYPE_UNSPECIFIED = 0;  // ENUM_TYPE_UNSPECIFIED
        ENUM_TYPE_CARRY_A     = 1;  // ENUM_TYPE_CARRY_A
        ENUM_TYPE_B           = 2;  // ENUM_TYPE_B
    :param flatten: if flatten nested object or not.
    Example of flatten:
      Original file:
         message Outer {
           int32 outerq = 1;
           message MiddleAA {
             message Inner {
               int64 ival = 1;
               bool  booly = 2;
             }
             bool  booly_aa = 1;
           }
         }

      Formatted file:
         message Inner {
             int64 ival = 1;
             bool booly = 2;
         }

         message MiddleAA {
             bool booly_aa = 1;
         }

         message Outer {
             int32 outerq = 1;
         }
    :param comment_max_length: the max length of comment line, defalut is no limitation.
    :return: formatted string.
    """
    parser = ProtoParser()
    protobuf_obj = parser.loads(proto_str=proto_str.strip())
    content = protobuf_obj.to_string(
        indents=indents,
        top_comment=top_comment,
        align_by_equal_sign=align_by_equal_sign,
        flatten=flatten,
        comment_max_length=comment_max_length
    )

    return content



from src.proto_formatter import format_str

proto_str = """
    /*
    Person balabala
*/
    message Person {
    // comment of name a
    // https:github.skyscannertools.net/fbi-partners/ndc-as-config/blob/master/engine/farelogix/templates/SeatAvailabilityRQ.xml#L18
required string name = 1; // comment of name b
/* 
comment of id a
// comment of id b
         */
        required int32 id = 2;// comment of id c
       optional string email = 3;// comment of email
}
"""
formatted_proto_str = format_str(proto_str, align_by_equal_sign=True, comment_max_length=120)
print(formatted_proto_str)


from enum import Enum

from proto_formatter.constant import Constant
from proto_formatter.comment import CommentParser
from proto_formatter.detector import Detector
from proto_formatter.import_parser import ImportParser
from proto_formatter.object_parser import ObjectParser
from proto_formatter.option_parser import OptionParser
from proto_formatter.package_parser import PackageParser
from proto_formatter.protobuf import Protobuf
from proto_formatter.syntax_parser import SyntaxParser
from proto_formatter.util import remove_prefix, remove_suffix


class ProtoParser(Constant):
    def __init__(self):
        self.protobuf_obj = Protobuf()

    def load(self, fp):
        """
        Load proto from file, deserialize it as a ProtoBufStructure object.
        :param fp: The absolute file path of the proto file.
        :return: Object of ProtoBufStructure.
        """
        lines = self._read_lines(fp)
        return self._parse(lines)

    def loads(self, proto_str):
        """
        Parse proto string, return a ProtoBufStructure object.
        :param proto_str: The proto string need to pasre.
        :return: Object of ProtoBufStructure.
        """
        lines = proto_str.split('\n')
        return self._parse(lines)

    def _read_lines(self, path):
        """
        Read data from file as line list, all blank and empty lines are removed before and after valid text.
        :param path: Absolute file path.
        :return: Line list of the striped file content.
        """
        with open(path) as f:
            content = f.read()
            content = content.strip()
            lines = content.split('\n')
            return lines

    def _parse(self, lines):
        """
        Pasre proto content lines, deserialize them as a ProtoBufStructure object.
        :param lines: the proto content lines.
        :return: an object of ProtoBufStructure.
        """
        top_comments = Detector().get_top_comments(lines)
        if len(lines) == 0:
            return self.protobuf_obj

        first_line = lines[0]
        type = Detector().get_type(first_line)

        parser = self._get_parser(type)
        if parser is None:
            return self.protobuf_obj

        if isinstance(parser, ObjectParser):
            parser.parse(lines)
            top_comments = CommentParser.create_comment(first_line, top_comments)
            all_comments = top_comments + parser.objects[0].comments
            parser.objects[0].comments = all_comments
            self.protobuf_obj.objects.extend(parser.objects)
        else:
            line = lines.pop(0)
            parser.parse_and_add(self.protobuf_obj, line, top_comments)

        if len(lines) == 0:
            return self.protobuf_obj

        return self._parse(lines)

    def _get_parser(self, keyword):
        if keyword == 'syntax':
            return SyntaxParser()
        elif keyword == 'package':
            return PackageParser()
        elif keyword == 'option':
            return OptionParser()
        elif keyword == 'import':
            return ImportParser()
        elif keyword == 'message':
            return ObjectParser()
        elif keyword == 'enum':
            return ObjectParser()
        elif keyword == 'service':
            return ObjectParser()
        elif keyword == 'oneof':
            return ObjectParser()



from proto_formatter.comment import CommentParser
from proto_formatter.protobuf import Protobuf
from proto_formatter.proto_structures import Syntax


class SyntaxParser():

    @classmethod
    def parse_and_add(cls, proto_obj: Protobuf, line, top_comment_list):
        if proto_obj.syntax is not None:
            raise 'multiple syntax detected!'

        proto_obj.syntax = cls.parse_syntax(line, top_comment_list)

    @classmethod
    def parse_syntax(cls, line, top_comment_list):
        value = cls._get_syntax_value(line)
        comments = CommentParser.create_comment(line, top_comment_list)
        syntax = Syntax(value, comments)
        return syntax

    @classmethod
    def _get_syntax_value(cls, line):
        line = line.strip().replace(' ', '')
        lindex = len('syntax=')
        rindex = line.index(';')
        value = line[lindex:rindex].strip().replace('"', "").replace("'", "")

        return value



from proto_formatter.proto_structures import Comment
from proto_formatter.constant import Constant
from proto_formatter.proto_structures import Position
from proto_formatter.util import remove_prefix, remove_suffix


class CommentParser(Constant):
    def __init__(self):
        self.end_line = 0

        self.multiple_comment_start_symbol_stack = []
        self.multiple_comment_end_symbol_stack = []

    def pick_up_comment(self, lines):
        comment_lines = []
        while lines:
            line = lines.pop(0).strip()

            if self._start_with_multiple_line_comment(line):
                self.multiple_comment_start_symbol_stack.append(self.MULTIPLE_COMENT_START_SYMBOL)
                comment_lines.append(line)
                continue

            if self.is_comment(line):
                comment_lines.append(line)

            if self._end_with_multiple_line_comment(line):
                self.multiple_comment_end_symbol_stack.append(self.MULTIPLE_COMENT_END_SYMBOL)
                continue

            if line and not self.is_comment(line):
                lines.insert(0, line)  # add back
                break

        return comment_lines

    @staticmethod
    def is_not_new_line(variable):
        return variable != '\n'

    def parse(self, comment_lines):
        text = '\n'.join(comment_lines)
        separator = '++++++++++++++++++'
        text = text.replace('/*', separator).replace('*/', separator)
        text_list = text.split(separator)
        text_list = list(filter(None, text_list))
        text_list = list(filter(self.is_not_new_line, text_list))
        text_list = [text.strip().replace(self.SINGLE_COMMENT_SYMBOL, '') for text in text_list]
        multiple_line_comments = text_list
        return multiple_line_comments

    def _start_with_single_line_comment(self, line):
        return line.strip().startswith(self.SINGLE_COMMENT_SYMBOL)

    def _start_with_multiple_line_comment(self, line):
        return line.strip().startswith(self.MULTIPLE_COMENT_START_SYMBOL)

    def _end_with_multiple_line_comment(self, line):
        return line.strip().endswith(self.MULTIPLE_COMENT_END_SYMBOL)

    def _is_multiple_comment(self):
        if len(self.multiple_comment_start_symbol_stack) == 0:
            return False

        return len(self.multiple_comment_start_symbol_stack) > len(self.multiple_comment_end_symbol_stack)

    def is_comment(self, line):
        if self._is_multiple_comment():
            return True

        if self._start_with_single_line_comment(line):
            return True

        return False

    @staticmethod
    def parse_single_line_comment(line):
        comment = None
        if line.count(CommentParser.SINGLE_COMMENT_SYMBOL) > 0:
            start_index = line.index(CommentParser.SINGLE_COMMENT_SYMBOL)
            comment_str = line[start_index:].lstrip(CommentParser.SINGLE_COMMENT_SYMBOL).strip()
            comment = Comment(comment_str, Position.Right)
        return comment

    @staticmethod
    def create_top_comments(comments):
        result = []
        while comments:
            comment_lines = comments.pop(0)
            text = ''.join(comment_lines)
            text = text.strip()
            text = remove_prefix(text, CommentParser.SINGLE_COMMENT_SYMBOL)
            text = remove_prefix(text, CommentParser.MULTIPLE_COMENT_START_SYMBOL)
            text = remove_suffix(text, CommentParser.MULTIPLE_COMENT_END_SYMBOL)
            text = text.strip()
            result.append(Comment(text, Position.TOP))
        return result

    @staticmethod
    def create_comment(line, top_comment_list):
        comments = CommentParser.create_top_comments(top_comment_list)
        right_comment = CommentParser.parse_single_line_comment(line)
        if right_comment is not None:
            comments.append(right_comment)
        return comments



from proto_formatter.comment import CommentParser
from proto_formatter.constant import Constant
from proto_formatter.proto_structures import Option
from proto_formatter.protobuf import Protobuf


class OptionParser(Constant):

    @classmethod
    def parse_and_add(cls, proto_obj: Protobuf, line, top_comment_list):
        option = cls.parse_option(line, top_comment_list)
        cls.add_options(proto_obj, option)

    @classmethod
    def add_options(cls, proto_obj: Protobuf, new_option):
        for option in proto_obj.options:
            if option.name == new_option.name:
                raise Exception(f'multiple option detected: {new_option.name}!')
        proto_obj.options.append(new_option)

    @classmethod
    def parse_option(cls, line, top_comment_list):
        name, value = cls._get_option_value(line)
        comments = CommentParser.create_comment(line, top_comment_list)
        option = Option(name, value, comments)
        return option

    @classmethod
    def _get_option_value(cls, line):
        line = line.strip()
        lindex = len('option ')
        equal_sign_index = line.index(cls.EQUAL_SIGN)
        semicolon_index = line.index(cls.SEMICOLON)
        name = line[lindex:equal_sign_index].strip().replace('"', "").replace("'", "")
        value = line[equal_sign_index + 1:semicolon_index].strip().replace('"', "").replace("'", "")

        return (name, value)
